//Network communication code
//Protocol version 3.0
//2x2-wire unidirectional buses
//tags and addresses supported
//Written by Kalim Moghul (kfm9 at cornell dot edu)

#include "ringbuffer.h"

//Pinout for bus communication
//Pin Signal    Direction
//7   BTx       IN
//6   BAck      IN
//5   PTx       OUT
//4   PAck      OUT
//3   DIn1      IN
//2   DIn0      IN
//1   DOut1     OUT
//0   DOut0     OUT

//PIDs are mapped to ports on the network switch as follows:
//Port      PID
//PORTA     0
//PORTB     1
//PORTC     2
//PORTD     3

#define BTx   0b10000000
#define BAck  0b01000000
#define PTx   0b00100000
#define PAck  0b00010000
#define Dmask 0b11111100

#define UBYTE unsigned char

//receive buffer size (actual data stored is 1 less since ring buffer needs a 1 byte sentinel)
#define RBUFSIZE 5

//this pragma ensures these globals won't go into registers
#pragma regalloc-
aos_semaphore *sem_net_tx;  /* Semaphore for writing from network */ 
aos_semaphore *sem_net_rx;  /* Semaphore for reading to network */ 

aos_semaphore *sem_net_rbuf;    /* Semaphore for reading from receive buffer */
UBYTE rbuf[16][RBUFSIZE];        /* allocate ringbuffer storage per tag/address pair */
ringbuffer rb[16];              /* allocate ringbuffer per tag/address pair */
#pragma regalloc+


//Initializes the network interface
//must be called at startup before switch starts reading
void net_init(void)
{
    int i;
    PORTC = 0x00;
    DDRC = 0x33;
    sem_net_tx = (aos_semaphore *)aos_sem_create( 1 );
    sem_net_rx = (aos_semaphore *)aos_sem_create( 1 );
    sem_net_rbuf = (aos_semaphore *)aos_sem_create( 1 );
    for (i=0; i < 16; i++)
        rbinit(rb + i, rbuf[i], RBUFSIZE);                 //init ring buffers
    //start listener thread here
}

#pragma regalloc-
static UBYTE txstate = 0, TAck = BAck;
#pragma regalloc+
//Performs a blocking send
//Parameters
//      Taddr = tag/address of recipient  (2-bits of tag and 2-bits of addr, 0bTTAA)
//      data = byte of data to be sent
void send_wait(UBYTE Taddr, UBYTE data)
{
    UBYTE notdone = 1;
    PORTA = ~(0x02);
    aos_wait(sem_net_tx);
    PORTA = ~(0x03);
    while (notdone)
    {
        if ((UBYTE)((UBYTE)PINC & (UBYTE)BAck) != (UBYTE)TAck)
        {
            TAck = TAck ^ BAck;
            if (txstate == 0)
            {
                PORTA = ~(0x04);
                PORTC = (UBYTE)((UBYTE)PORTC & (UBYTE)Dmask) | (UBYTE)((UBYTE)Taddr & (UBYTE)0x3);            //put addr crumb
                txstate++;
            }
            else if (txstate == 1)
            {
                PORTA = ~(0x05);
                PORTC = (UBYTE)((UBYTE)PORTC & (UBYTE)Dmask) | (UBYTE)((UBYTE)(Taddr >> 2) & (UBYTE)0x3);    //put tag crumb
                txstate++;
            }
            else
            {
                PORTC = (UBYTE)((UBYTE)PORTC & (UBYTE)Dmask) | (UBYTE)((UBYTE)(data >> ((5 - txstate) << 1)) & (UBYTE)0x3);   //put data crumb, MSBs go first
                txstate = ((txstate >= 5) ? 0 : txstate + 1);       //do 4 iterations of data
                if (txstate == 0) notdone = 0;
            }
            PORTC = (UBYTE)PORTC ^ (UBYTE)PTx;                                  //toggle the transmit signal
        }
    }
    aos_signal(sem_net_tx);
}

#pragma regalloc-
static UBYTE rxstate = 0, RAck = 0;
#pragma regalloc+
//Performs a blocking recieve
//Modifies
//      global rb[(tag<<2)|addr] will contain the received data
void receive_wait(void)
{
    UBYTE notdone = 1;
    UBYTE Raddr, Rtag, Rdata;
    aos_wait(sem_net_rx);
    while (notdone)
    {
        if ((UBYTE)((UBYTE)PINC & (UBYTE)BTx) != (UBYTE)RAck)                               //if signal was toggled
        {
            RAck = (UBYTE)RAck ^ (UBYTE)BTx;                                    //store the last Ack state (toggle)
            if (rxstate == 0)
            {
                Raddr = (UBYTE)(PINC >> 2) & (UBYTE)0x3;                        //read address crumb (2-bits)
                PORTC = (UBYTE)PORTC ^ (UBYTE)PAck;                             //toggle the Ack
                rxstate++;
            }
            else if (rxstate == 1)
            {
                Rtag = (UBYTE)(PINC >> 2) & (UBYTE)0x3;                         //read tag crumb
                PORTC = (UBYTE)PORTC ^ (UBYTE)PAck;                             //toggle the Ack
                rxstate++;
            }
            else
            {
                Rdata = (UBYTE)(Rdata << 2) | (UBYTE)((UBYTE)(PINC >> 2) & (UBYTE)0x3);         //read data crumb
                rxstate = ((rxstate >= 5) ? 0 : rxstate + 1);                   //do 4 iterations of data
                PORTC = (UBYTE)PORTC ^ (UBYTE)PAck;                             //toggle the Ack                
                if (rxstate == 0)                                               //finished receiving, store data
                {
                    notdone = 0;
                    while(1)    //stall here while trying to find space in the ring buffer
                    {
                        aos_wait(sem_net_rbuf);
                        if(rbput(rb + ((Rtag<<2)|Raddr), Rdata))                //add to ring buffer
                        {
                            aos_signal(sem_net_rbuf);
                            break;
                        }
                        aos_signal(sem_net_rbuf);
                    }
                }
            }
        }
    }
    aos_signal(sem_net_rx);
}

//Performs a blocking send of a buffer
//Parameters
//      addr = address of target
//      tag = tag to receive data
//      length = number of bytes to transmit from buffer
//      buf = pointer to buffer containing message
void net_send(UBYTE addr, UBYTE tag, UBYTE length, UBYTE *buf)
{
    UBYTE i, Taddr;
    Taddr = (tag << 2) | (addr & 0x3);
    for (i = 0; i < length; i++)
        send_wait(Taddr, buf[i]);
}

//Performs a blocking receive from a specific source
//Parameters
//      addr = address of source to receive from (255=any)
//      tag = local tag to receive data from (255=any)
//      length = maximum number of bytes to receive
//Returns
//      The number of bytes actually received
//Modifies
//      buf contains the received data
UBYTE net_recv(UBYTE addr, UBYTE tag, UBYTE length, UBYTE *buf)
{
    UBYTE count = 0, *temp, st, et, sa, ea, i, j;
    aos_wait(sem_net_rbuf);

    //determine tag and source address ranges to get data from
    if ((tag == 255) && (addr == 255))      //any tag, any source
    {
        st = 0; et = 3;
        sa = 0; ea = 3;
    }
    else if (tag == 255)                    //any tag, one source
    {
        st = 0; et = 3;
        sa = ea = addr;
    }
    else if (addr == 255)                   //one tag, any source
    {
        st = et = tag;
        sa = 0; ea = 3;
    }
    else                                    //one tag, one source
    {
        st = et = tag;
        sa = ea = addr;
    }
    for (i = st; (i <= et) && (count < length); i++)
    {
        for (j = sa; (j <= ea) && (count < length); j++)
        {
            do
            {
                temp = rbget(rb + ((i<<2)|j));
                if (temp != NULL)
                {
                    buf[count] = *temp;
                    count++;
                }
            }
            while ((count < length) && temp);
        }
    }
    aos_signal(sem_net_rbuf);
    return count;
}